#! /usr/bin/perl -w
#$Id: server.pl,v 1.35 2004/06/07 04:44:41 mstorti Exp $ 

use strict;
use English;
use Getopt::Std;
my %opts;
getopts("N:n:hsqw:r:b:k:d",\%opts);

if ($opts{'h'}) {
/`/;
    print <<'EOM';
usage:  $ server.pl [OPTIONS]

OPTIONS: -h                 print help
         -N <number>        number of rounds
         -n <number>        print each n rounds 
         -s                 stop after printing each board
         -q                 don't print board, statistics only
         -w <filename>      the file with the world description (def: tiny2.world)
         -r <filename>      the file with the brain of the red ant (def: ant.txt)
         -b <filename>      the file with the brain of the black ant (def: ant.txt)
         -k <number>        keep a certain number of records of ant history
                               (0 keeps all, default: no keep)
         -d                 dump ant commands as they are executed
EOM
/`/;
    exit;
}
my $quiet = $opts{'q'};
my $keep_history = $opts{'k'};
my $dumpa = $opts{'d'};

my $s = 12345;
my ($NX,$NY);
my @board;
my @ANTS = ();
my ($red_ant_brain, $black_ant_brain);

my ($STATE,$COLOR,$RESTING,$DIRECTION,$HAS_FOOD,$X,$Y,$HISTORY) = (0..7);
my ($ROCKY,$ANT,$FOOD,$MARKERS,$HILL) = (0..4);
my ($LEFT,$RIGHT) = (0..1);

sub randomint {
    my $p = shift();
    $s = ($s * 22695477)% 1073741824 + 1;
    return (int($s/65536) % 16384) % $p;
}

# for (my $j=0; $j<20; $j++) {
#     my $x = randomint();
#     print "x $x\n";
# }

## usage @ngbr = adjacent_cell(\@pos,$d);
sub adjacent_cell {
    my ($pos,$d) = @_;
    my ($x,$y) = @$pos;
    return ($d==0? ($x+1,$y) :
	    $d==1? (!($y % 2) ? ($x,$y+1) : ($x+1,$y+1)) :
	    $d==2? (!($y % 2) ? ($x-1,$y+1) : ($x,$y+1)) :
	    $d==3? ($x-1,$y) :
	    $d==4? (!($y % 2) ? ($x-1,$y-1) : ($x,$y-1)) :
	    $d==5? (!($y % 2) ? ($x,$y-1) : ($x+1,$y-1)) :
	    die "bud d: $d\n");
}

my $L=-1;
my $R=+1;

# $newdir = turn($lr,$olddir)
sub turn {
    my ($lr,$d) = @_;
    return ($lr==$LEFT ? ($d-1)% 6 : $lr==$RIGHT ? ($d+1)% 6 : 
	    die("bad turn dir\n"));
}

# $pos2 = sensed_cell($pos,$d,$sd);
sub sensed_cell {
    my ($pos,$d,$sd) = @_;
    my ($x,$y) = @$pos;
    if ($sd eq 'here') { return @$pos; }
    elsif ($sd eq 'ahead') { return adjacent_cell($pos,$d); }
    elsif ($sd eq 'leftahead') { return adjacent_cell($pos,turn($d,$L)); }
    elsif ($sd eq 'rightahead') { return adjacent_cell($pos,turn($d,$R)); }
    else { die "bad sd: $sd\n"; }
}

my ($RED,$BLACK)=(0,1);
sub other_color {
    my $c = shift();
    return 1-$c;
}

## @ant = (state,color,resting,direction,has_food,x,y);
## usage $cell = cell(\@pos);
## @$cell = ($rocky,$ant,$food,$markers,$hill);
## @$markers = (0,1,11...)  (length 12)
sub cell {
    my $pos = shift();
    my ($x,$y) = @$pos;
    return $board[$y][$x];
}

## function ant_is_alive(id:int):bool
sub ant_is_alive {
    my $id = shift();
    my $ant = $ANTS[$id];
    return $ant->[5]>=0;
}

sub find_ant {
    my $id = shift();
    my $ant = $ANTS[$id];
    return ($ant->[5],$ant->[6]);
}

sub kill_ant_at{
    my $pos = @_;
    my $cell = cell($pos);
    my $ant = $cell->[1];
    if ($ant!=undef) {
	$cell->[1] = undef;
	$ANTS[$ant]->[5]=-1;
    }
}
	
sub food_at {
    my $cell = cell(shift());
    return $cell->[$FOOD];
}

sub set_food_at {
    my ($pos,$f) = @_;
    my $cell = cell($pos);
    $cell->[$FOOD] = $f;
}

sub anthill_at() {
    my ($pos,$c) = @_;
    my $cell = cell($pos);
    return ($cell->[4]==$c);
}


sub read_board {
    my $name = shift();
    open BOARD,$name;
    $NX = <BOARD>;
    $NY = <BOARD>;
    chomp $NX;
    chomp $NY;
    for (my $j=0; $j<$NY; $j++) {
	my $line = <BOARD>;
	chomp $line;
	my @line = split " ",$line;
    ##	print "line $j, NX $NX, chars",scalar @line,"\n";
	push @board,[@line];
    }
    close BOARD;

## @$cell = ($rocky,$ant,$food,$markers,$hill);
## @ant = (state,color,resting,direction,has_food,x,y);
    my @nomarks = (0) x 12;
    my $id=0;
    for (my $j=0; $j<$NY; $j++) {
	for (my $k=0; $k<$NX; $k++) {
	    my $c = $board[$j][$k];
	    if ($c eq '#') { $board[$j][$k] = [1,undef,0,undef,undef]; }
	    elsif ($c eq '.') { $board[$j][$k] = [0,undef,0,[@nomarks],undef]; }
	    elsif ($c eq '+') { 
		$board[$j][$k] = [0,$id++,0,[@nomarks],$RED]; 
		push @ANTS,[0,$RED,0,0,0,$k,$j,[]];
	    }
	    elsif ($c eq '-') { 
		$board[$j][$k] = [0,$id++,0,[@nomarks],$BLACK]; 
		push @ANTS,[0,$BLACK,0,0,0,$k,$j,[]];
	    }
	    elsif ($c =~ /\d/) { $board[$j][$k] = [0,undef,$c,[@nomarks],undef]; }
	    else              { print "Bad char reading board: \"$c\"\n"; }
	}
    }
    print $#ANTS+1," ants initialized\n";
}

# function set_marker_at(p:pos, c:color, i:marker) =
# set_marker_at($pos,$c,$i,$val)
sub set_marker_at {
    my ($pos,$c,$i,$val) = @_;
    my $cell = cell($pos);
    my $markers = $cell->[3];
    $markers->[$c+$i]=$val;
}

# function check_any_marker_at(p:pos, c:color):bool =
sub check_any_marker_at {
    my ($pos,$c) = @_;
    my $cell = cell($pos);
    my $markers = $cell->[3];
    for (my $j=0; $j<6; $j++) { 
	return 1 if $markers->[$c*6+$j];
    }
    return 0;
}

#function cell_matches(p:pos, cond:condition, c:color):bool = 
## @$cell = ($rocky,$ant,$food,$markers,$hill);
## @ant = (state,color,resting,direction,has_food,x,y,history);

# bool cell_matches($pos,$cond,$c,$marker);
sub cell_matches {
    my ($pos,$cond,$c,$marker) = @_;
    my $cell = cell($pos);
    if ($cell->[$ROCKY]) { return ($cond eq 'rock' ? 1 : 0); }
    my $ant = $cell->[$ANT];
    if ($cond eq 'friend') {
	return defined $ant && $ANTS[$ant]->[1]==$c;
    } elsif ($cond eq 'foe') {
	return defined $ant && $ANTS[$ant]->[1]==(1-$c);
    } elsif ($cond eq 'friendwithfood') {
	return defined $ant && $ANTS[$ant]->[1]==$c
	    && $ANTS[$ant]->[4];
    } elsif ($cond eq 'foewithfood') {
	return defined $ant && $ANTS[$ant]->[1]==(1-$c)
	    && $ANTS[$ant]->[4];
    } elsif ($cond eq 'food') {
	return $cell->[$FOOD]>0;
    } elsif ($cond eq 'rock') {
	return 0;
    } elsif ($cond eq 'marker') {
	my $markers = $cell->[3];
	return $markers->[$c+$marker];
    } elsif ($cond eq 'foemarker') {
	return check_any_marker_at($pos,1-$c);
    } elsif ($cond eq 'home') {	
	return defined $cell->[4] && $cell->[4]==$c;
    } elsif ($cond eq 'foehome') {	
	return defined $cell->[4] && $cell->[4]==(1-$c);
    } else { die "Bad cond \"$cond\"\n"; }
}

sub read_ant {
    my $name = shift();
    die "ant brain file: \"$name\" doesn't exist\n" 
	unless -f $name;
    open ANT,$name 
	|| die "Can't open ant brain file: \"$name\", reason: $!\n";
    my @ant = ();
    while (1) {
	my $line = <ANT>;
	my $readline = $line;
	last unless $line;
	chomp $line;
	$line = $PREMATCH if $line =~ /\s*;.*$/;
	$line = lc($line);
	my @cmd=();
	while ($line =~ /\s*(\S+)/) { 
	    push @cmd,$1; 
	    $line = $POSTMATCH;
	}
	die unless $#cmd>=0;
	chomp $readline;
	push @ant,[@cmd,$readline];
    }
    close ANT;
    print $#ant+1," comands in $name\n";
    return [@ant];
}

# function adjacent_ants(p:pos, c:color):int =
sub adjacent_ants {
    my ($pos,$c) = @_;
    my $n=0;
    my $cell = cell($pos);
    my $ant = $cell->[1];
    for (my $d=0; $d<6; $d++) {
	$n++ if defined  $ant && $ANTS[$ant]->[1]==$c;
    }
    return $n;
}

# function check_for_surrounded_ant_at(p:pos) =
sub check_for_surrounded_ant_at {
    my $pos = shift();
    my $cell = cell($pos);
    my $antid = $cell->[1];
    return 0 unless defined $antid;
    my $ant = $ANTS[$antid];
    my $c = $ant->[1];
    if (adjacent_ants($pos,1-$c)>=5) {
	my $ant_food = ($ant->[4] ? 1 : 0);
	kill_ant_at($pos);
	$cell->[$FOOD] += 3 + $ant_food;
    }
}

sub check_for_surrounded_ants {
    my $pos = shift();
    for (my $d=0; $d<6; $d++) {
	my @pos = adjacent_cell($pos,$d);
	check_for_surrounded_ant_at(\@pos);
    }
}

sub keep_history {
    my ($hist,$state) = @_;
    push @$hist,$state;
    shift @$hist 
	if $keep_history>0 && @$hist>$keep_history;
}

## @ant = (state,color,resting,direction,has_food,x,y);
## @$cell = ($rocky,$ant,$food,$markers,$hill);
sub step {
    my $id = shift();
    return unless ant_is_alive($id);
    my $ant = $ANTS[$id];
    my $pos = [$ant->[5],$ant->[6]];
    my $cell = cell($pos);
    
    if ($ant->[$RESTING]>0) { $ant->[$RESTING]--; }
    else {
	my $color = $ant->[$COLOR];
	my $state = $ant->[$STATE];
	my $brain = ($color == $RED ? $red_ant_brain :
		     $color == $BLACK ? $black_ant_brain :
		     die "unknown color \"$color\"\n");
	my $hist = $ant->[$HISTORY];
	if (!defined $state || $state>=@$brain || $state<0) {
	    print "last states: \n";
	    foreach my $state (@$hist) {
		my $instr = $brain->[$state];
		my $line = $instr->[-1];
		print "$line\n";
	    }
	    die "bad state \n";
	}
	keep_history($hist,$state) if defined $keep_history;
	if ($dumpa) {
	    my $instr = $brain->[$state];
	    my $line = $instr->[-1];
	    print "ant $id: $line\n";
	}
	my $instr_ref = $brain->[$state];
	my @instr = @$instr_ref;
	my $instr = \@instr;
	my $cmd = shift @instr;
#	print "$state: ",join (" ",@$instr_ref),"\n";
	if ($cmd eq 'sense') {
	    my @pos2 = sensed_cell($pos,$ant->[$DIRECTION],$instr->[0]);
	    my $cond = $instr->[3];
	    my $marker = $instr->[4];
	    my $newstate = (cell_matches(\@pos2,$cond,$color,$marker) ?
			    $instr->[1] : $instr->[2]);
	    $ant->[$STATE] = $newstate;
	} elsif ($cmd eq 'mark') {
	    set_marker_at($pos,$color,$instr->[0],1);
	    $ant->[$STATE] = $instr->[1];
	} elsif ($cmd eq 'unmark') {
	    set_marker_at($pos,$color,$instr->[0],0);
	    $ant->[$STATE] = $instr->[1];
	} elsif ($cmd eq 'pickup') {
	    if ($ant->[$HAS_FOOD] || $cell->[$FOOD]==0) {
		 $ant->[$STATE] = $instr->[1];
	     } else {
		 $cell->[$FOOD]--;
		 $ant->[$HAS_FOOD] = 1;
		 $ant->[$STATE] = $instr->[0];
	     }
	} elsif ($cmd eq 'drop') {
	    if ($ant->[$HAS_FOOD]) {
		$cell->[$FOOD]++;
		$ant->[$HAS_FOOD] = 0;
	    }
	    $ant->[$STATE] = $instr->[0];
	} elsif ($cmd eq 'turn') {
	    my $olddir = $ant->[$DIRECTION];
	    my $LR = $instr->[0];
	    my $lr = ($LR eq 'right' ? $RIGHT 
		      : $LR eq 'left' ? $LEFT 
		      : die("Bad lr: $LR\n"));
	    my $newdir = turn($lr,$olddir);
	    $ant->[$DIRECTION] = $newdir;
	    $ant->[$STATE] = $instr->[1];
	} elsif ($cmd eq 'move') {
	    my $dir = $ant->[$DIRECTION];
	    my @newpos = adjacent_cell($pos,$dir);
	    my $newcell = cell(\@newpos);
	    if ($newcell->[$ROCKY] || defined $newcell->[$ANT]) {
		$ant->[$STATE] = $instr->[1];
	    } else {
		$cell->[$ANT]=undef;
		$newcell->[$ANT] = $id;
		$ant->[$X] = $newpos[0];
		$ant->[$Y] = $newpos[1];
		$ant->[$RESTING] = 14;
		check_for_surrounded_ants(\@newpos);
		$ant->[$STATE] = $instr->[0];
	    }
	} elsif ($cmd eq 'flip') {
	    my $p = $instr->[0];
	    $ant->[$STATE] = (randomint($p) ? 
			      $instr->[2] : $instr->[1]);
	} else { 
	    die "Bad command \"$cmd\"\n"; 
	}
    }
}

my $print_mark_as_value = $RED;
$print_mark_as_value = undef;

my $vis_mode = 1;
my ($ccs,$rocky,$shift);
if ($vis_mode==1) {
    $ccs = 4; ## cell size
    $ccs += 1 if defined $print_mark_as_value;
    $rocky = "[".('#' x $ccs)."]";
    $shift = " " x int(($ccs+2)/2);
} elsif ($vis_mode==2) {
    $ccs = 1;
    $rocky = "#";
    $shift="";
} else { die "unknown vis. mode \"$vis_mode\"\n"}

sub print_board {
    my $round = shift();
    $round  = "<unknown>" unless defined $round;
    my ($food_red,$food_black,$rants,$bants) = (0,0,0,0);
    for (my $j=0; $j<$NY; $j++) {
	if ($j % 2) { print $shift unless $quiet; }
	for (my $k=0; $k<$NX; $k++) {
	    my @pos = ($k,$j);
	    my $cell = cell(\@pos);
	    if ($cell->[$ROCKY]) { print $rocky unless $quiet; }
	    else {
		my $hill = $cell->[$HILL];
		if (defined $hill) {
		    if ($hill==$RED) { $food_red += $cell->[$FOOD]; }
		    elsif ($hill==$BLACK) { $food_black += $cell->[$FOOD]; }
		}
		my $char='.';
		my $id = $cell->[$ANT];
		my @s;
		my $f = $cell->[$FOOD];
		$s[0] = ($f==0? '.' : $f>=10 ? 'F' : $f);
		if (defined $id) {
		    my $ant = $ANTS[$id];
		    my $c = $ant->[$COLOR];
		    if ($c == $RED) { $rants++; }
		    elsif ($c == $BLACK) { $bants++; }
		    $s[1] = ($c == $RED ? "r" : "b");
		    if ($ant->[$HAS_FOOD]) { $s[1] = uc($s[1]); }
		} else { $s[1] = ' '; }
		my $markers = $cell->[$MARKERS];
		my $m = 0;
		if (! defined $print_mark_as_value) {
		    for (my $j=0; $j<12; $j++) { 
			if ($markers->[$j]) { $m=1; last; }
		    }
		    $s[2] = ($m ? 'M' : ' ');
		} else {
		    my $val = 0;
		    for (my $j=0; $j<6; $j++) { 
			my $c = $print_mark_as_value;
			$val = 2*$val + $markers->[$c*6+$j];
		    }
# 		    if ($val>0) {
# 			print "mrk: ",join("",@$markers),
# 			", val: $val\n";
# 		    }
		    $s[2] = sprintf "%02d",$val;
		}
		$s[3] = (!defined $hill ? ' ' : $hill==$RED ? '+' : '-');
		if ($vis_mode==1) {
		    print '[',@s,']' unless $quiet;
		} elsif ($vis_mode==2) {
		    $char = $s[1];
		    print $char unless $quiet;
		} else { die "unknown vis. mode \"$vis_mode\"\n"}
	    }
	}
	print "\n" unless $quiet;
    }
    print "$round: food in hills: ${food_red}R, ${food_black}B, ",
    "ants: ${rants}R, ${bants}B\n\n";
}

my $world = "tiny2.world";
$world = $opts{'w'} if defined $opts{'w'};

my $rant = "ant.txt";
$rant = $opts{'r'} if defined $opts{'r'};

my $bant = "ant.txt";
$bant = $opts{'b'} if defined $opts{'b'};

read_board($world);
$red_ant_brain = read_ant($rant);
$black_ant_brain = read_ant($bant);

my $nround = 10000;
$nround = $opts{'N'} if $opts{'N'};
my $nprint = 100;
$nprint = $opts{'n'} if $opts{'n'};
for (my $round=0; $round<$nround; $round++) {
    for (my $id=0; $id<@ANTS; $id++) { step($id); }
    if (!($round % $nprint)) { 
	print "round: $round\n";
	print_board($round); 
	if ($opts{'s'}) {
	    my $cmd = <STDIN>;
	    eval($cmd) unless $cmd =~ /^\s*$/;
	}
    }
}

